/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  Users and possessors of this source code 
 * are hereby granted a nonexclusive, royalty-free license to use this code 
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.   This source code is a "commercial item" as 
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer  software"  and "commercial computer software 
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein. 
 *
 * Any use of this source code in individual and commercial software must 
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

#include <stdio.h>
#include <cutil_inline.h>

#define BIN_COUNT 256
#define HISTOGRAM_SIZE (BIN_COUNT * sizeof(unsigned int))

////////////////////////////////////////////////////////////////////////////////
// GPU-specific definitions
////////////////////////////////////////////////////////////////////////////////
//Machine warp size
#define WARP_LOG2SIZE 5

//Warps in thread block for histogram256Kernel()
#define WARP_N 12

//Corresponding thread block size in threads for histogram256Kernel()
#define THREAD_N (WARP_N << WARP_LOG2SIZE)

//Total histogram size (in counters) per thread block for histogram256Kernel()
#define BLOCK_MEMORY (WARP_N * BIN_COUNT)

//Thread block count for histogram256Kernel()
#define BLOCK_N 64

//#define D_S 528 // Pal - 192 threads/block
//#define D_S 396 // Pal - 256 threads/block
#define D_S 264 // Pal - 384 threads/block
//#define D_S 132 // Half-pal
//#define D_S 33 // Quarter-pal

////////////////////////////////////////////////////////////////////////////////
// Shortcut shared memory atomic addition functions
////////////////////////////////////////////////////////////////////////////////
__device__ inline void addData256(unsigned int *s_WarpHist, unsigned int data){
        atomicAdd(s_WarpHist + data, 1);
}

////////////////////////////////////////////////////////////////////////////////
// Main histogram calculation kernel
////////////////////////////////////////////////////////////////////////////////
static __global__ void histogram256Kernel(unsigned int *d_Result, unsigned int *d_Data, int dataN){
    //Current global thread index
    const int    globalTid = blockIdx.x * blockDim.x + threadIdx.x;
    //Total number of threads in the compute grid
    const int   numThreads = blockDim.x * gridDim.x;
    const int H = (int)(blockIdx.x / D_S);

    //Shared memory storage for each warp
    __shared__ unsigned int s_Hist[BLOCK_MEMORY];

    //Current warp shared memory base
    const int warpBase = (threadIdx.x >> WARP_LOG2SIZE) * BIN_COUNT;

    //Clear shared memory buffer for current thread block before processing
    for(int pos = threadIdx.x; pos < BLOCK_MEMORY; pos += blockDim.x)
       s_Hist[pos] = 0;

    //Cycle through the entire data set, update subhistograms for each warp
    __syncthreads();

    for(int pos = globalTid; pos < dataN; pos += numThreads){
        unsigned int data4 = d_Data[pos];

        addData256(s_Hist + warpBase, (data4 >>  0) & 0xFFU);
        addData256(s_Hist + warpBase, (data4 >>  8) & 0xFFU);
        addData256(s_Hist + warpBase, (data4 >> 16) & 0xFFU);
        addData256(s_Hist + warpBase, (data4 >> 24) & 0xFFU);
    }

    __syncthreads();
    //Merge per-warp histograms into per-block and write to global memory
    for(int pos = threadIdx.x; pos < BIN_COUNT; pos += blockDim.x){
        unsigned int sum = 0;
        for(int base = 0; base < BLOCK_MEMORY; base += BIN_COUNT)
            sum += s_Hist[base + pos] & 0x07FFFFFFU;
            atomicAdd(d_Result + H*BIN_COUNT + pos, sum);
    }
}
