/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CUDA streams - JGL - 10/11/2010
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Optimally streamed 256-bin histogram calculation
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// streams_Histo256_Optimal.cu
// This file contains the code used in Section 4.4. It is based on histogram256 included in CUDA SDK.
// Pixel distribution changes in a given time, in order to recalculate the optimum.
//
// Reference:
// @article{GomezLuna20121117,
//   title = "Performance models for asynchronous data transfers on consumer Graphics Processing Units",
//   journal = "Journal of Parallel and Distributed Computing",
//   volume = "72",
//   number = "9",
//   pages = "1117 - 1126",
//   year = "2012",
//   note = "<ce:title>Accelerators for High-Performance Computing</ce:title>",
//   issn = "0743-7315",
//   doi = "10.1016/j.jpdc.2011.07.011",
//   url = "http://www.sciencedirect.com/science/article/pii/S0743731511001468",
//   author = "Juan G\'omez-Luna and Jos\'e Mar\'ia Gonz\'alez-Linares and Jos\'e Ignacio Benavides and Nicol\'as Guil",
// }
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>

////////////////////////////////////////////////////////////////////////////////
// Common host and device functions
////////////////////////////////////////////////////////////////////////////////
//Round a / b to nearest higher integer value
int iDivUp(int a, int b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}
//Round a / b to nearest lower integer value
int iDivDown(int a, int b){
    return a / b;
}
//Align a to nearest higher multiple of b
int iAlignUp(int a, int b){
    return (a % b != 0) ?  (a - a % b + b) : a;
}
//Align a to nearest lower multiple of b
int iAlignDown(int a, int b){
    return a - a % b;
}

int aprox_equal(float a, float b){
	if((a/b <= 1.20) && (a/b >= 0.80)) return(1);
	else return(0);
}

////////////////////////////////////////////////////////////////////////////////
// Reference CPU histogram
////////////////////////////////////////////////////////////////////////////////
extern "C" void histogram256CPU(unsigned int *h_ResultCPU, unsigned int *h_Data, int data_N);

////////////////////////////////////////////////////////////////////////////////
// Include kernel
////////////////////////////////////////////////////////////////////////////////
#include <histogram256_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// Data configuration
////////////////////////////////////////////////////////////////////////////////
const int DATA_W = iAlignUp(704, 16);
const int DATA_H = 576;
const int TOTAL_FRAMES = 4096;
const int DATA_SIZE = DATA_W * DATA_H;
const int DATA_SIZE4 = DATA_W * DATA_H / 4;
const int DATA_SIZE_INT = DATA_SIZE4 * sizeof(unsigned int);

//Carry out dummy calculations before main computation loop
//in order to "warm up" the hardware/driver
#define WARMUP

////////////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv){

	unsigned int *h_DataA;
	unsigned char *hh_DataA, *hhh_DataA;
	unsigned int *h_histoCPU, *h_histoGPU;

	unsigned char *d_DataA;
	unsigned int *d_histo;

	double sum_delta2,sum_ref2,L1norm2;
	unsigned int hTimer;
	float elapsed_time, time_memcpy1, time_memcpy2, time_kernel;   // timing variables
	float ELAPSED_TIME, TIME_MEMCPY1, TIME_MEMCPY2, TIME_KERNEL, TIME_ESTIMATE, TIME_STREAM, TIME_NONSTREAMED, TIME_STREAMED, TIME_OPTIMIZED;   // timing variables
	ELAPSED_TIME=0; TIME_MEMCPY1=0; TIME_MEMCPY2=0; TIME_KERNEL=0; TIME_NONSTREAMED=0; TIME_STREAMED=0; TIME_OPTIMIZED=0;
	int i, x;
	int nreps = 10;

	cutCreateTimer(&hTimer);

	// Check the compute capability of the device
	int num_devices=0;
	cudaGetDeviceCount(&num_devices);
	if(0==num_devices){
		printf("Your system does not have a CUDA capable device\n");
		return 1;
	}
	cudaDeviceProp device_properties;
	cudaGetDeviceProperties(&device_properties,0);
	if((1==device_properties.major)&&(device_properties.minor<1))
	printf("%s does not have compute capability 1.1 or later\n\n",device_properties.name);

	// Syntax verification
	if (argc != 3) {
		printf("Wrong format\n");
		printf("Syntax: %s <nstreams> <FRAMES>\n",argv[0]);
		exit(1);
	}
	int nstreams = atoi(argv[1]);
	int FRAMES = atoi(argv[2]);

	// GTX 580
	float t_sc = 0.01;
	// GTX 480
	//float t_sc = 0.03;
	// GTX 280
	//float t_sc = 0.1;
	printf("Device %s\tDATA_W %d\tTOTAL_FRAMES %d\tFRAMES %d\tnstreams %d\n", device_properties.name, DATA_W, TOTAL_FRAMES, FRAMES, nstreams);

	// Host memory allocation
	cudaMallocHost((void **)&h_DataA, DATA_SIZE_INT*TOTAL_FRAMES);
	cudaMallocHost((void **)&hh_DataA, DATA_SIZE*sizeof(unsigned char)*FRAMES);
	cudaMallocHost((void **)&hhh_DataA, DATA_SIZE*sizeof(unsigned char)*TOTAL_FRAMES);
	cudaMallocHost((void **)&h_histoCPU, HISTOGRAM_SIZE*TOTAL_FRAMES);
	cudaMallocHost((void **)&h_histoGPU, HISTOGRAM_SIZE*TOTAL_FRAMES);

	// Device memory allocation
	cudaMalloc((void **)&d_DataA, DATA_SIZE_INT*FRAMES);
	cudaMalloc((void **)&d_histo, HISTOGRAM_SIZE*FRAMES);

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Input images generation
	///////////////////////////////////////////////////////////////////////////////////////////////////
	srand(time(NULL));
	for(i = 0; i < DATA_SIZE*FRAMES; i++)
		hh_DataA[i] = (unsigned char)(rand() % 256);

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef WARMUP
	//printf("Warm up...\n");
	cudaMemcpy(d_DataA+0*DATA_SIZE, hh_DataA+0*DATA_SIZE, DATA_SIZE_INT, cudaMemcpyHostToDevice);
	histogram256Kernel<<<BLOCK_N, THREAD_N>>>(d_histo+0*BIN_COUNT, (unsigned int *)(d_DataA+0*DATA_SIZE), DATA_SIZE/4);
	cudaThreadSynchronize();
#endif
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Create CUDA event handles
	cudaEvent_t start_event,stop_event;
	cudaEventCreate(&start_event);
	cudaEventCreate(&stop_event);

	////////////////////////////////////// INPUT GENERATION /////////////////////////////////////////
	// Half random - half degenerate
	srand(time(NULL));
	for(i = 0; i < DATA_SIZE*TOTAL_FRAMES/2; i++)
		hhh_DataA[i] = (unsigned char)(rand() % 256);
	for(i = DATA_SIZE*TOTAL_FRAMES/2; i < DATA_SIZE*TOTAL_FRAMES; i++)
		hhh_DataA[i] = (unsigned char)0;
	memcpy(hh_DataA, hhh_DataA+0*(TOTAL_FRAMES/FRAMES)*DATA_SIZE, DATA_SIZE_INT*FRAMES);
	memcpy(h_DataA, hhh_DataA, DATA_SIZE_INT*TOTAL_FRAMES);

	////////////////////////////////////// RANDOM /////////////////////////////////////////
	// First test: measuring execution time for two types of frames
	cudaEventRecord(start_event,0);     // Transfer random frames
	for (x = 0; x < nreps; ++x){
		cudaMemcpy(d_DataA, hh_DataA, DATA_SIZE_INT*FRAMES, cudaMemcpyHostToDevice);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy1,start_event,stop_event);
	TIME_MEMCPY1 += time_memcpy1 / nreps;

	cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
	cudaEventRecord(start_event,0);
	for (x = 0; x < nreps; ++x){
		histogram256Kernel<<<DATA_SIZE4*FRAMES/THREAD_N, THREAD_N>>>(d_histo, (unsigned int *)d_DataA, FRAMES*DATA_SIZE/4);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_kernel,start_event,stop_event);
	TIME_KERNEL += time_kernel / nreps;

	cudaEventRecord(start_event,0);
	for (x = 0; x < nreps; ++x){     // Transfer histograms
		cudaMemcpy(h_histoGPU,d_histo,FRAMES*HISTOGRAM_SIZE,cudaMemcpyDeviceToHost);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy2,start_event,stop_event);
	TIME_MEMCPY2 += time_memcpy2 / nreps;

	printf("Random: CPU-GPU %f\tKernel %f\tGPU-CPU %f\n", TIME_MEMCPY1, TIME_KERNEL, TIME_MEMCPY2);

	printf("Random:\t%f\n", TIME_MEMCPY1 + TIME_KERNEL + TIME_MEMCPY2);	

	////////////////////////////////////// DEGENERATE /////////////////////////////////////////
	memcpy(hh_DataA, hhh_DataA+(TOTAL_FRAMES/2)*DATA_SIZE, DATA_SIZE_INT*FRAMES);
	TIME_MEMCPY1 = 0; TIME_KERNEL = 0;
	// First test: measuring execution time for two types of frames
	cudaEventRecord(start_event,0);     // Transfer degenerate frame
	for (x = 0; x < nreps; ++x){
		cudaMemcpy(d_DataA, hh_DataA, DATA_SIZE_INT*FRAMES, cudaMemcpyHostToDevice);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy1,start_event,stop_event);
	TIME_MEMCPY1 += time_memcpy1 / nreps;

	cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
	cudaEventRecord(start_event,0);
	for (x = 0; x < nreps; ++x){
		histogram256Kernel<<<DATA_SIZE4*FRAMES/THREAD_N, THREAD_N>>>(d_histo, (unsigned int *)d_DataA, FRAMES*DATA_SIZE/4);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_kernel,start_event,stop_event);
	TIME_KERNEL += time_kernel / nreps;
	
	printf("Degenerate: CPU-GPU %f\tKernel %f\tGPU-CPU %f\n", TIME_MEMCPY1, TIME_KERNEL, TIME_MEMCPY2);	
	
	printf("Degenerate:\t%f\n\n", TIME_MEMCPY1 + TIME_KERNEL + TIME_MEMCPY2);

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// NON-STREAMED EXECUTION //////////////////////////////////////////////
	memset(h_histoGPU, 0, HISTOGRAM_SIZE*TOTAL_FRAMES);
	TIME_MEMCPY1 = 0; TIME_KERNEL = 0; TIME_MEMCPY2 = 0;

	int z = 0;
	while(z < TOTAL_FRAMES / FRAMES){	
		cudaEventRecord(start_event,0);
        cudaMemcpy(d_DataA, hhh_DataA+z*FRAMES*DATA_SIZE, DATA_SIZE_INT*FRAMES, cudaMemcpyHostToDevice);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&TIME_MEMCPY1,start_event,stop_event);

		cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
		cudaEventRecord(start_event,0);
		histogram256Kernel<<<DATA_SIZE4*FRAMES/THREAD_N, THREAD_N>>>(d_histo, (unsigned int *)d_DataA, FRAMES*DATA_SIZE/4);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&TIME_KERNEL,start_event,stop_event);

		cudaEventRecord(start_event,0);
       	cudaMemcpy(h_histoGPU+z*FRAMES*BIN_COUNT, d_histo, FRAMES*HISTOGRAM_SIZE, cudaMemcpyDeviceToHost);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&TIME_MEMCPY2,start_event,stop_event);

		TIME_NONSTREAMED += TIME_MEMCPY1 + TIME_KERNEL + TIME_MEMCPY2;
		z++;
	}
	printf("z = %d TIME_NOSTREAM = %f\n", z, TIME_NONSTREAMED);


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// STREAMED EXECUTION //////////////////////////////////////////////
	memset(h_histoGPU, 0, HISTOGRAM_SIZE*TOTAL_FRAMES);
	TIME_MEMCPY1 = 0; TIME_KERNEL = 0; TIME_MEMCPY2 = 0;
	z = 0;

	cudaStream_t stream[nstreams];
	for (i = 0; i < nstreams; ++i) cudaStreamCreate(&stream[i]);

	dim3 BLOCK_S((DATA_SIZE4/THREAD_N)*(FRAMES/nstreams));

	while(z < TOTAL_FRAMES / FRAMES){	
			// Streamed computation
			cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
			cudaEventRecord(start_event,0);

			for (i = 0; i < nstreams; ++i)
				cudaMemcpyAsync(d_DataA+i*DATA_SIZE*FRAMES/nstreams, hhh_DataA+z*FRAMES*DATA_SIZE+i*DATA_SIZE*FRAMES/nstreams, DATA_SIZE_INT*FRAMES/nstreams, cudaMemcpyHostToDevice, stream[i]);

			for (i = 0; i < nstreams; ++i){
				histogram256Kernel<<<BLOCK_S, THREAD_N, 0, stream[i]>>>(d_histo+i*BIN_COUNT*FRAMES/nstreams, (unsigned int *)(d_DataA+i*DATA_SIZE*FRAMES/nstreams), (DATA_SIZE/4)*FRAMES/nstreams);
			}
	       	for (i = 0; i < nstreams; ++i)
			cudaMemcpyAsync(h_histoGPU+z*FRAMES*BIN_COUNT+i*BIN_COUNT*FRAMES/nstreams, d_histo+i*BIN_COUNT*FRAMES/nstreams, HISTOGRAM_SIZE*FRAMES/nstreams, cudaMemcpyDeviceToHost, stream[i]);

			cudaEventRecord(stop_event,0);
			cudaEventSynchronize(stop_event);
			cudaEventElapsedTime(&TIME_STREAM,start_event,stop_event);

			z++;
			TIME_STREAMED += TIME_STREAM;
	}
	for(int i = 0; i < nstreams; i++)
		cudaStreamDestroy(stream[i]);
	printf("z = %d nstreams = %d TIME_STREAMED = %f\n", z, nstreams, TIME_STREAMED);


	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// OPTIMALLY STREAMED EXECUTION ////////////////////////////////////////
	memset(h_histoGPU, 0, HISTOGRAM_SIZE*TOTAL_FRAMES);
	TIME_MEMCPY1 = 0; TIME_KERNEL = 0; TIME_MEMCPY2 = 0;

	z = 0;
	while(z < TOTAL_FRAMES / FRAMES){	
	/////////////////////////////////////////////////////////////////////////////////////////////////////NON-STREAMED
		// Determining the optimum after a sequential execution
		cudaEventRecord(start_event,0);
        cudaMemcpy(d_DataA, hhh_DataA+z*FRAMES*DATA_SIZE, DATA_SIZE_INT*FRAMES, cudaMemcpyHostToDevice);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&TIME_MEMCPY1,start_event,stop_event);

		cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
		cudaEventRecord(start_event,0);
		histogram256Kernel<<<DATA_SIZE4*FRAMES/THREAD_N, THREAD_N>>>(d_histo, (unsigned int *)d_DataA, FRAMES*DATA_SIZE/4);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&TIME_KERNEL,start_event,stop_event);

		cudaEventRecord(start_event,0);
       	cudaMemcpy(h_histoGPU+z*FRAMES*BIN_COUNT, d_histo, FRAMES*HISTOGRAM_SIZE, cudaMemcpyDeviceToHost);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&TIME_MEMCPY2,start_event,stop_event);

		TIME_OPTIMIZED += TIME_MEMCPY1 + TIME_KERNEL + TIME_MEMCPY2;
		z++;

		// Stream management: determining the value of nstreams with the previous measures
		if(TIME_MEMCPY1 > TIME_KERNEL){
			nstreams = (int)sqrt(TIME_KERNEL / t_sc);
			printf("nstreams - Transf.Dom. %f %d\t--->", sqrt(TIME_KERNEL / t_sc), nstreams);
		}
		else{
			nstreams = (int)sqrt(TIME_MEMCPY1 / t_sc);
			printf("nstreams - Kernel.Dom. %f %d\t--->", sqrt(TIME_MEMCPY1 / t_sc), nstreams);
		}

		if(nstreams == 1) nstreams = 2;
		while(FRAMES%nstreams != 0) nstreams--;
		//while(FRAMES%nstreams != 0) nstreams++;
		printf("nstreams = %d\t--->\t", nstreams);

		if(TIME_MEMCPY1 > TIME_KERNEL){
			TIME_ESTIMATE = TIME_MEMCPY1 + (TIME_KERNEL / nstreams) + TIME_MEMCPY2 + nstreams * t_sc;
			printf("TIME_ESTIMATE = %f\n", TIME_ESTIMATE);
		}
		else{
			TIME_ESTIMATE = (TIME_MEMCPY1 / nstreams) + TIME_KERNEL + TIME_MEMCPY2 + nstreams * t_sc;
			printf("TIME_ESTIMATE = %f\n", TIME_ESTIMATE);
		}
		// Stream management: determining the value of nstreams with the previous measures - FOR C.C. 1.X
		/*if(TIME_MEMCPY1 + TIME_MEMCPY2 > TIME_KERNEL){
			nstreams = (int)((TIME_MEMCPY1+TIME_MEMCPY2)/(TIME_MEMCPY1+TIME_MEMCPY2-TIME_KERNEL));
			printf("nstreams - Transf.Dom. %f %d\t--->", (TIME_MEMCPY1+TIME_MEMCPY2)/(TIME_MEMCPY1+TIME_MEMCPY2-TIME_KERNEL), nstreams);
		}
		else{
			nstreams = (int)sqrt((TIME_MEMCPY1+TIME_MEMCPY2)/ t_sc);
			printf("nstreams - Kernel.Dom. %f %d\t--->", sqrt((TIME_MEMCPY1+TIME_MEMCPY2) / t_sc), nstreams);
		}
		//while(FRAMES%nstreams != 0) nstreams--;
		while(FRAMES%nstreams != 0) nstreams++;
		printf("nstreams = %d\t--->\t", nstreams);
		if(TIME_MEMCPY1 + TIME_MEMCPY2 > TIME_KERNEL){
			TIME_ESTIMATE = TIME_MEMCPY1 + TIME_MEMCPY2 + nstreams * t_sc;
			printf("TIME_ESTIMATE = %f\n", TIME_ESTIMATE);
		}
		else{
			TIME_ESTIMATE = ((TIME_MEMCPY1 + TIME_MEMCPY2)/ nstreams) + TIME_KERNEL + nstreams * t_sc;
			printf("TIME_ESTIMATE = %f\n", TIME_ESTIMATE);
		}*/
		
		TIME_STREAM = TIME_ESTIMATE;

		cudaStream_t stream[nstreams];
		for (i = 0; i < nstreams; ++i) cudaStreamCreate(&stream[i]);

		dim3 BLOCK_S((DATA_SIZE4/THREAD_N)*(FRAMES/nstreams));

		while((aprox_equal(TIME_STREAM, TIME_ESTIMATE)) && (z < TOTAL_FRAMES/FRAMES)){
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////STREAM
			// Streamed computation using the optimum
			cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);

			cudaEventRecord(start_event,0);
			for (i = 0; i < nstreams; ++i)
				cudaMemcpyAsync(d_DataA+i*DATA_SIZE*FRAMES/nstreams, hhh_DataA+z*FRAMES*DATA_SIZE+i*DATA_SIZE*FRAMES/nstreams, DATA_SIZE_INT*FRAMES/nstreams, cudaMemcpyHostToDevice, stream[i]);

			for (i = 0; i < nstreams; ++i){
				histogram256Kernel<<<BLOCK_S, THREAD_N, 0, stream[i]>>>(d_histo+i*BIN_COUNT*FRAMES/nstreams, (unsigned int *)(d_DataA+i*DATA_SIZE*FRAMES/nstreams), (DATA_SIZE/4)*FRAMES/nstreams);
			}
	       	for (i = 0; i < nstreams; ++i)
				cudaMemcpyAsync(h_histoGPU+z*FRAMES*BIN_COUNT+i*BIN_COUNT*FRAMES/nstreams, d_histo+i*BIN_COUNT*FRAMES/nstreams, HISTOGRAM_SIZE*FRAMES/nstreams, cudaMemcpyDeviceToHost, stream[i]);
			cudaEventRecord(stop_event,0);
			cudaEventSynchronize(stop_event);
			cudaEventElapsedTime(&TIME_STREAM,start_event,stop_event);

			z++;
			TIME_OPTIMIZED += TIME_STREAM;
		}

		for(int i = 0; i < nstreams; i++)
			cudaStreamDestroy(stream[i]);

	}
	printf("z = %d TIME_OPTIMALLY_STREAMED = %f\n", z, TIME_OPTIMIZED);
	

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// CPU computing
	///////////////////////////////////////////////////////////////////////////////////////////////////
	memset(h_histoCPU, 0, HISTOGRAM_SIZE*TOTAL_FRAMES);
	cudaEventRecord(start_event,0);
	for(i = 0; i < TOTAL_FRAMES; ++i){
		histogram256CPU(h_histoCPU+i*BIN_COUNT, (unsigned int *)h_DataA+i*DATA_SIZE4, DATA_SIZE/4);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&elapsed_time,start_event,stop_event);
	ELAPSED_TIME += elapsed_time;
	printf("CPU time %f\n", ELAPSED_TIME);

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Comparing the results
	///////////////////////////////////////////////////////////////////////////////////////////////////
        sum_delta2 = 0;
        sum_ref2   = 0;
	L1norm2 = 0;
        for(int i = 0; i < BIN_COUNT*TOTAL_FRAMES; i++){
            sum_delta2 += fabs(h_histoCPU[i] - h_histoGPU[i]);
            sum_ref2   += fabs(h_histoCPU[i]);
        }
        L1norm2 = sum_delta2 / sum_ref2;
	//printf("L1norm2 B1: %E\n", L1norm2);
	printf((L1norm2 < 1e-6) ? "PASSED\n" : "FAILED\n");

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	///////////////////////////////////////////////////////////////////////////////////////////////////
	cudaEventDestroy(start_event);
	cudaEventDestroy(start_event);
	// Device memory
        cudaFree(d_DataA);
        cudaFree(d_histo);
	// Host memory
        cudaFreeHost(h_DataA);
        cudaFreeHost(hh_DataA);
        cudaFreeHost(hhh_DataA);
        cudaFreeHost(h_histoCPU);
        cudaFreeHost(h_histoGPU);

	cutDeleteTimer(hTimer);

}

///////////////////////////////////////////////////////////////////////////////////////////////////
