/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CUDA streams - JGL - 10/11/2010
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Streamed matrix multiplication
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// streams_Matrix.cu
// This file contains the code used in Section 4.1. It computes a matrix multiplication C = A * B.
// It is based on matrix multiplication included in CUDA SDK.
//
// Reference:
// @article{GomezLuna20121117,
//   title = "Performance models for asynchronous data transfers on consumer Graphics Processing Units",
//   journal = "Journal of Parallel and Distributed Computing",
//   volume = "72",
//   number = "9",
//   pages = "1117 - 1126",
//   year = "2012",
//   note = "<ce:title>Accelerators for High-Performance Computing</ce:title>",
//   issn = "0743-7315",
//   doi = "10.1016/j.jpdc.2011.07.011",
//   url = "http://www.sciencedirect.com/science/article/pii/S0743731511001468",
//   author = "Juan G\'omez-Luna and Jos\'e Mar\'ia Gonz\'alez-Linares and Jos\'e Ignacio Benavides and Nicol\'as Guil",
// }
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Utilities and system includes
#include "cutil_inline.h"

// includes, kernels
#include <matrixMul_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest(int argc, char** argv);
void randomInit(float*, int);

extern "C"
void computeGold(float*, const float*, const float*, unsigned int, unsigned int, unsigned int);

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
    cudaError_t error = cudaSetDevice(0);
    if (error != cudaSuccess) {
        printf("Error setting device to %d: %s\n", 1, cudaGetErrorString(error));
    }


    runTest(argc, argv);

}

////////////////////////////////////////////////////////////////////////////////
// Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
void runTest(int argc, char** argv)
{
	// Syntax verification
	if (argc != 2) {
		printf("Wrong format\n");
		printf("Syntax: %s <nstreams>\n",argv[0]);
		exit(1);
	}
	int nstreams = atoi(argv[1]);   // number of streams for CUDA calls
		
	float elapsed_time, time_memcpy1, time_memcpy2, time_kernel;   // timing variables

	// allocate and initialize an array of stream handles
	cudaStream_t *streams = (cudaStream_t*) malloc(nstreams * sizeof(cudaStream_t));
	for(int i = 0; i < nstreams; i++)
		cudaStreamCreate(&(streams[i]));

	// create CUDA event handles
	cudaEvent_t start_event, stop_event;
	cudaEventCreate(&start_event);
	cudaEventCreate(&stop_event);

    int nIter = 100;

    int devID;
    cudaDeviceProp props;

    // get number of SMs on this GPU
    cudaGetDevice(&devID);
    cudaGetDeviceProperties(&props, devID);

    printf("Device %d: \"%s\"\t", devID, props.name, props.major, props.minor);

	// set seed for rand()
    srand(2006);

    // Optional Command-line multiplier for matrix sizes
    unsigned int uiWA, uiHA, uiWB, uiHB, uiWC, uiHC;
    int iSizeMultiple = 1;

	// For GPUs with fewer # of SM's, we limit the maximum size of the matrix
	if (props.multiProcessorCount <= 4) {
		uiWA = 2 * BLOCK_SIZE * iSizeMultiple;
		uiHA = 4 * BLOCK_SIZE * iSizeMultiple;
		uiWB = 2 * BLOCK_SIZE * iSizeMultiple;
		uiHB = 4 * BLOCK_SIZE * iSizeMultiple;
		uiWC = 2 * BLOCK_SIZE * iSizeMultiple;
		uiHC = 4 * BLOCK_SIZE * iSizeMultiple;
	} else {
		uiWA = WA * iSizeMultiple;
		uiHA = HA * iSizeMultiple;
		uiWB = WB * iSizeMultiple;
		uiHB = HB * iSizeMultiple;
		uiWC = WC * iSizeMultiple;
		uiHC = HC * iSizeMultiple;
	}
    printf("Matrix Sizes: A( %u x %u ), B( %u x %u ), C( %u x %u )\n", 
            uiWA, uiHA, uiWB, uiHB, uiWC, uiHC);

    // allocate host memory for matrices A and B
    unsigned int size_A = uiWA * uiHA;
    unsigned int mem_size_A = sizeof(float) * size_A;
    float* h_A;
    cudaMallocHost((void **)&h_A, mem_size_A);
    unsigned int size_B = uiWB * uiHB;
    unsigned int mem_size_B = sizeof(float) * size_B;
    float* h_B;
    cudaMallocHost((void **)&h_B, mem_size_B);

    // initialize host memory
    randomInit(h_A, size_A);
    randomInit(h_B, size_B);

    // allocate device memory
    float* d_A;
    cudaMalloc((void**) &d_A, mem_size_A);
    float* d_B;
    cudaMalloc((void**) &d_B, mem_size_B);

    // allocate device memory for result
    unsigned int size_C = uiWC * uiHC;
    unsigned int mem_size_C = sizeof(float) * size_C;
    float* d_C;
    cudaMalloc((void**) &d_C, mem_size_C);

    // allocate host memory for the result
    float* h_C;
    cudaMallocHost((void **)&h_C, mem_size_C);
    float* h_C2;
    cudaMallocHost((void **)&h_C2, mem_size_C);

    // Matrix B is previously transferred
	cudaMemcpy(d_B, h_B, mem_size_B, cudaMemcpyHostToDevice);

	////////////////////////////////////// TIME MEMCOPY HOST TO DEVICE /////////////////////////////////////////
	// time memcopy to device - SYNCHRONOUS
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nIter; k++){
		// copy host memory to device
		cudaMemcpy(d_A, h_A, mem_size_A, cudaMemcpyHostToDevice);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy1, start_event, stop_event);
	printf("CPU-GPU SYNC:\t%.2f\n", time_memcpy1/nIter);

	//////////////////////////////////////////// TIME KERNEL ////////////////////////////////////////////////////
	// time kernel - SYNCHRONOUS
	// setup execution parameters
	dim3 threads(BLOCK_SIZE, BLOCK_SIZE);
	dim3 grid(uiWC / threads.x, uiHC / threads.y);

	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nIter; k++){
		// kernel warmup
		matrixMul<<< grid, threads >>>(d_C, d_A, d_B, uiWA, uiWB);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_kernel, start_event, stop_event);
	printf("kernel - SYNC:\t%.2f\n", time_kernel/nIter);

	////////////////////////////////////// TIME MEMCOPY DEVICE TO HOST /////////////////////////////////////////
	// time memcopy from device - SYNCHRONOUS
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nIter; k++){
		// copy result from device to host
		cudaMemcpy(h_C, d_C, mem_size_C, cudaMemcpyDeviceToHost);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy2, start_event, stop_event);
	printf("GPU-CPU SYNC:\t%.2f\n", time_memcpy2/nIter);

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// NON-STREAMED EXECUTION //////////////////////////////////////////////
	// time non-streamed execution for reference
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nIter; k++){
		cudaMemcpy(d_A, h_A, mem_size_A, cudaMemcpyHostToDevice);
        matrixMul<<< grid, threads >>>(d_C, d_A, d_B, uiWA, uiWB);
		cudaMemcpy(h_C2, d_C, mem_size_C, cudaMemcpyDeviceToHost);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&elapsed_time, start_event, stop_event);
	printf("non-streamed:\t%.2f (%.2f expected)\tblocks = %d x %d, threads = %d x %d\n", elapsed_time / nIter, time_memcpy1/nIter + time_kernel/nIter + time_memcpy2/nIter, grid.x, grid.y, threads.x, threads.y);

    cudaThreadSynchronize();

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// STREAMED EXECUTION //////////////////////////////////////////////////
	// time execution with nstreams streams
	dim3 threadsS(BLOCK_SIZE, BLOCK_SIZE);
	dim3 blocksS(uiWC / threads.x, (uiHC / threads.y) / nstreams);
	cudaMemset(d_C, 0, mem_size_C);
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nIter; k++){
		for(int i = 0; i < nstreams; i++)
			cudaMemcpyAsync(d_A + i * size_A / nstreams, h_A + i * size_A / nstreams, mem_size_A / nstreams, cudaMemcpyHostToDevice, streams[i]);

		for(int i = 0; i < nstreams; i++)
		        matrixMul<<< blocksS, threadsS, 0, streams[i] >>>(d_C + i * size_C / nstreams, d_A + i * size_A / nstreams, d_B, uiWA, uiWB);

		for(int i = 0; i < nstreams; i++)
			cudaMemcpyAsync(h_C + i * size_C / nstreams, d_C + i * size_C / nstreams, mem_size_C / nstreams, cudaMemcpyDeviceToHost, streams[i]);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&elapsed_time, start_event, stop_event);
	printf("%d streams:\t%.2f (%.2f expected with compute capability 1.1 or later)\tblocks = %d x %d, threads = %d x %d\n", 
		nstreams, elapsed_time / nIter, time_kernel/nIter + (time_memcpy1/nIter+time_memcpy2/nIter) / nstreams, blocksS.x, blocksS.y, threadsS.x, threadsS.y);


    // compute reference solution
	// printf("\nCheck against Host computation...\n\n");    
	float* reference = (float*)malloc(mem_size_C);
	// computeGold(reference, h_A, h_B, uiHA, uiWA, uiWB);

   double sum_delta2 = 0;
   double sum_ref2   = 0;
   double L1norm2 = 0;
   for(int i = 0; i < size_C; i++){
           sum_delta2 += fabs(h_C2[i] - reference[i]);
           sum_ref2   += fabs(reference[i]);
   }
   L1norm2 = (double)(sum_delta2 / sum_ref2);
   //printf("L1norm2: %E, sum_delta2: %f, sum_ref2: %f\n", L1norm2, sum_delta2, sum_ref2);
   //printf((L1norm2 < 1e-6) ? "PASSED\n" : "FAILED\n");

   sum_delta2 = 0;
   sum_ref2   = 0;
   L1norm2 = 0;
   for(int i = 0; i < size_C; i++){
           sum_delta2 += fabs(h_C[i] - h_C2[i]);
           sum_ref2   += fabs(h_C2[i]);
   }
   L1norm2 = (double)(sum_delta2 / sum_ref2);
   //printf("L1norm2: %E, sum_delta2: %f, sum_ref2: %f\n", L1norm2, sum_delta2, sum_ref2);
   printf((L1norm2 < 1e-6) ? "PASSED\n" : "FAILED\n");

    // clean up memory
    cudaFreeHost(h_A);
    cudaFreeHost(h_B);
    cudaFreeHost(h_C);
    cudaFreeHost(h_C2);
    free(reference);
    cudaFree(d_A);
    cudaFree(d_B);
    cudaFree(d_C);

    cudaThreadExit();
}

// Allocates a matrix with random float entries.
void randomInit(float* data, int size)
{
    for (int i = 0; i < size; ++i)
        data[i] = rand() / (float)RAND_MAX;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
