/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CUDA streams - JGL - 10/11/2010
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Characterizing the behavior of CUDA streams - Second observations: asynchronous transfers
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// streams_Method_SecondObs.cu
// This file contains the code used in Section 3.1.2. Variable data transfers size from CPU to GPU and back.
// It is based on simpleStreams.cu included in CUDA SDK.
//
// Reference:
// @article{GomezLuna20121117,
//   title = "Performance models for asynchronous data transfers on consumer Graphics Processing Units",
//   journal = "Journal of Parallel and Distributed Computing",
//   volume = "72",
//   number = "9",
//   pages = "1117 - 1126",
//   year = "2012",
//   note = "<ce:title>Accelerators for High-Performance Computing</ce:title>",
//   issn = "0743-7315",
//   doi = "10.1016/j.jpdc.2011.07.011",
//   url = "http://www.sciencedirect.com/science/article/pii/S0743731511001468",
//   author = "Juan G\'omez-Luna and Jos\'e Mar\'ia Gonz\'alez-Linares and Jos\'e Ignacio Benavides and Nicol\'as Guil",
// }
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <cutil.h>
#include <signal.h>
#include <time.h>

__global__ void init_array(int *g_data, int *factor, int num_iterations){ 

	int idx = blockIdx.x * blockDim.x + threadIdx.x;

	for(int i=0;i<num_iterations;i++)
		g_data[idx] += *factor;
}

int main(int argc, char *argv[]){
	int cuda_device = 0;
	// Syntax verification
	if (argc != 4) {
		printf("Wrong format\n");
		printf("Syntax: %s <nstreams niterations data_CPU-GPU>\n",argv[0]);
		exit(1);
	}
	int nstreams = atoi(argv[1]);   // number of streams for CUDA calls
	int niterations = atoi(argv[2]);   // number of iterations within the kernel
	int data1 = atoi(argv[3]);	// data size
	int data2 = 1024*1024;	// 1M
	int n1 = data1*data2;       // number of ints in the data set - CPU to GPU
	int n1bytes = n1 * sizeof(int);   // number of data bytes
	int n2 = (24-data1)*data2;       // number of ints in the data set - GPU to CPU
	int n2bytes = n2 * sizeof(int);   // number of data bytes
	int Nbytes = 24*data2*sizeof(int);	// device memory allocated
	int m = 24*data2;
	int nreps = 50;                 // number of times each experiment is repeated

	dim3 threads, blocks;           // kernel launch configuration

	float elapsed_time, time_memcpy1, time_memcpy2, time_kernel;   // timing variables

	// check the compute capability of the device
	int num_devices=0;
	cudaGetDeviceCount(&num_devices);
	if(0==num_devices){
		printf("your system does not have a CUDA capable device\n");
		return 1;
	}
	cudaDeviceProp device_properties;
	cudaGetDeviceProperties(&device_properties, cuda_device);
	if( (1 == device_properties.major) && (device_properties.minor < 1))
		printf("%s does not have compute capability 1.1 or later\n\n", device_properties.name);
	cudaSetDevice(cuda_device);

	printf("> Device name : %s\t", device_properties.name );
	//printf("> CUDA Capable SM %d.%d hardware with %d multi-processors\n", device_properties.major, device_properties.minor, device_properties.multiProcessorCount); 
	printf("array_size   = %d\t%d\t%d\t%d\n", n1, n2, niterations, data1);

	// allocate host memory
	int c = 5;                      // value to which the array will be initialized
	int *a = 0;                     // pointer to the array data in host memory
	// allocate host memory (pinned is required for achieve asynchronicity)
	cudaMallocHost((void**)&a, Nbytes);
	for(int i=0; i<m; i++) a[i] = 1;

	// allocate device memory
	int *d_a = 0, *d_c = 0;             // pointers to data and init value in the device memory
	cudaMalloc((void**)&d_a, Nbytes);
	cudaMemset(d_a, 0, Nbytes);
	cudaMalloc((void**)&d_c, sizeof(int));
	cudaMemcpy(d_c, &c, sizeof(int), cudaMemcpyHostToDevice);

	// allocate and initialize an array of stream handles
	cudaStream_t *streams = (cudaStream_t*) malloc(nstreams * sizeof(cudaStream_t));
	for(int i = 0; i < nstreams; i++)
		cudaStreamCreate(&(streams[i]));

	// create CUDA event handles
	cudaEvent_t start_event, stop_event;
	cudaEventCreate(&start_event);
	cudaEventCreate(&stop_event);

	////////////////////////////////////// TIME MEMCOPY HOST TO DEVICE /////////////////////////////////////////
	// time memcopy to device - SYNCHRONOUS
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nreps; k++){
		cudaMemcpy(d_a, a, n1bytes, cudaMemcpyHostToDevice);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy1, start_event, stop_event);
	printf("CPU-GPU SYNC:\t%.2f\n", time_memcpy1/nreps);

	//////////////////////////////////////////// TIME KERNEL ////////////////////////////////////////////////////
	// time kernel - SYNCHRONOUS
	threads=dim3(768, 1);
	blocks=dim3(m / threads.x, 1);
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nreps; k++){
		init_array<<<blocks, threads>>>(d_a, d_c, niterations);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_kernel, start_event, stop_event);
	printf("kernel - SYNC:\t%.2f\n", time_kernel/nreps);

	////////////////////////////////////// TIME MEMCOPY DEVICE TO HOST /////////////////////////////////////////
	// time memcopy from device - SYNCHRONOUS
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nreps; k++){
		cudaMemcpy(a, d_a, n2bytes, cudaMemcpyDeviceToHost);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&time_memcpy2, start_event, stop_event);
	printf("GPU-CPU SYNC:\t%.2f\n", time_memcpy2/nreps);

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// NON-STREAMED EXECUTION //////////////////////////////////////////////
	// time non-streamed execution for reference
	blocks=dim3(m / threads.x, 1);
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nreps; k++){
		cudaMemcpy(d_a, a, n1bytes, cudaMemcpyHostToDevice);
		init_array<<<blocks, threads>>>(d_a, d_c, niterations);
		cudaMemcpy(a, d_a, n2bytes, cudaMemcpyDeviceToHost);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&elapsed_time, start_event, stop_event);
	printf("non-streamed:\t%.2f (%.2f expected)\tblocks = %d, threads = %d\n", elapsed_time / nreps, time_memcpy1/nreps + time_kernel/nreps + time_memcpy2/nreps, blocks.x, threads.x);

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// STREAMED EXECUTION //////////////////////////////////////////////////
	// time execution with nstreams streams
	blocks=dim3(m/(nstreams*threads.x),1);
	memset(a, 0, Nbytes);
	cudaMemset(d_a, 0, Nbytes);
	cudaEventRecord(start_event, 0);
	for(int k = 0; k < nreps; k++){
		// asynchronously launch nstreams memcopies
		for(int i = 0; i < nstreams; i++)
			cudaMemcpyAsync(d_a + i * n1 / nstreams, a + i * n1 / nstreams, n1bytes / nstreams, cudaMemcpyHostToDevice, streams[i]);

		// asynchronously launch nstreams kernels, each operating on its own portion of data
		for(int i = 0; i < nstreams; i++){
			init_array<<<blocks, threads, 0, streams[i]>>>(d_a + i * n1 / nstreams, d_c, niterations);
		}

		// asynchronously launch nstreams memcopies
		for(int i = 0; i < nstreams; i++)
			cudaMemcpyAsync(a + i * n2 / nstreams, d_a + i * n2 / nstreams, n2bytes / nstreams, cudaMemcpyDeviceToHost, streams[i]);
	}
	cudaEventRecord(stop_event, 0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&elapsed_time, start_event, stop_event);
	printf("%d streams:\t%.2f (%.2f expected with compute capability 1.1 or later)\tblocks = %d, threads = %d\n", nstreams, elapsed_time / nreps, time_kernel/nreps + (time_memcpy1/nreps+time_memcpy2/nreps) / nstreams, blocks.x, threads.x);

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// release resources
	for(int i = 0; i < nstreams; i++)
		cudaStreamDestroy(streams[i]);
	cudaEventDestroy(start_event);
	cudaEventDestroy(stop_event);
	cudaFreeHost(a);
	cudaFree(d_a);
	cudaFree(d_c);

	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
