/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CUDA streams - JGL - 10/11/2010
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Streamed RGB to grayscale
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// streams_RGB2gray.cu
// This file contains the code used in Section 4.3. It generates a grayscale histogram per frame.
// It is based on histogram256 included in CUDA SDK.
//
// Reference:
// @article{GomezLuna20121117,
//   title = "Performance models for asynchronous data transfers on consumer Graphics Processing Units",
//   journal = "Journal of Parallel and Distributed Computing",
//   volume = "72",
//   number = "9",
//   pages = "1117 - 1126",
//   year = "2012",
//   note = "<ce:title>Accelerators for High-Performance Computing</ce:title>",
//   issn = "0743-7315",
//   doi = "10.1016/j.jpdc.2011.07.011",
//   url = "http://www.sciencedirect.com/science/article/pii/S0743731511001468",
//   author = "Juan G\'omez-Luna and Jos\'e Mar\'ia Gonz\'alez-Linares and Jos\'e Ignacio Benavides and Nicol\'as Guil",
// }
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>

////////////////////////////////////////////////////////////////////////////////
// Common host and device functions
////////////////////////////////////////////////////////////////////////////////
//Round a / b to nearest higher integer value
int iDivUp(int a, int b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}
//Round a / b to nearest lower integer value
int iDivDown(int a, int b){
    return a / b;
}
//Align a to nearest higher multiple of b
int iAlignUp(int a, int b){
    return (a % b != 0) ?  (a - a % b + b) : a;
}
//Align a to nearest lower multiple of b
int iAlignDown(int a, int b){
    return a - a % b;
}

////////////////////////////////////////////////////////////////////////////////
// Reference CPU histogram
////////////////////////////////////////////////////////////////////////////////
extern "C" void rgb2grayCPU(unsigned int *h_ResultCPU, unsigned int *h_Data, int data_N);

////////////////////////////////////////////////////////////////////////////////
// Include kernel
////////////////////////////////////////////////////////////////////////////////
#include <streams_RGB2gray_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// Data configuration
////////////////////////////////////////////////////////////////////////////////
const int DATA_W = iAlignUp(176, 16);
const int DATA_H = 144;
const int FRAMES = 32;
const int DATA_SIZE = DATA_W * DATA_H;
const int DATA_SIZE_INT = DATA_SIZE * sizeof(unsigned int);

//Carry out dummy calculations before main computation loop
//in order to "warm up" the hardware/driver
#define WARMUP

////////////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv){

	unsigned int *h_DataA;
	unsigned char *hh_DataA;
	unsigned int *h_histoCPU, *h_histoGPU, *h_histoGPU1, *h_histoGPU2;

	unsigned char *d_DataA;
	unsigned int *d_histo;

	double sum_delta2,sum_ref2,L1norm2;
	unsigned int hTimer;
	float elapsed_time, time_memcpy1, time_memcpy2, time_kernel;   // timing variables
	float time_stream, time_nostream;
	float ELAPSED_TIME, TIME_MEMCPY1, TIME_MEMCPY2, TIME_KERNEL;   // timing variables
	float TIME_stream, TIME_nostream;
	ELAPSED_TIME=0; TIME_MEMCPY1=0; TIME_MEMCPY2=0; TIME_KERNEL=0;
	TIME_stream=0; TIME_nostream=0;
	int i, x;
	int nreps = 5;

	cutCreateTimer(&hTimer);

	// Check the compute capability of the device
	int num_devices=0;
	cudaGetDeviceCount(&num_devices);
	if(0==num_devices){
		printf("Your system does not have a CUDA capable device\n");
		return 1;
	}
	cudaDeviceProp device_properties;
	cudaGetDeviceProperties(&device_properties,0);
	if((1==device_properties.major)&&(device_properties.minor<1))
	printf("%s does not have compute capability 1.1 or later\n\n",device_properties.name);

	// Host memory allocation
	cudaMallocHost((void **)&h_DataA, DATA_SIZE_INT*FRAMES);
	cudaMallocHost((void **)&hh_DataA, 4*DATA_SIZE*sizeof(unsigned char)*FRAMES);
	cudaMallocHost((void **)&h_histoCPU, HISTOGRAM_SIZE*FRAMES);
	cudaMallocHost((void **)&h_histoGPU, HISTOGRAM_SIZE*FRAMES);
	cudaMallocHost((void **)&h_histoGPU1, HISTOGRAM_SIZE*FRAMES);
	cudaMallocHost((void **)&h_histoGPU2, HISTOGRAM_SIZE*FRAMES);

	// Device memory allocation
	cudaMalloc((void **)&d_DataA, DATA_SIZE_INT*FRAMES);
	cudaMalloc((void **)&d_histo, HISTOGRAM_SIZE*FRAMES);

	// Syntax verification
	if (argc != 2) {
		printf("Wrong format\n");
		printf("Syntax: %s <nstreams>\n",argv[0]);
		exit(1);
	}
	int nstreams = atoi(argv[1]);

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Input images generation
	///////////////////////////////////////////////////////////////////////////////////////////////////
	srand(2007);
	for(i = 0; i < 4*DATA_SIZE*FRAMES; i++)
		hh_DataA[i] = (unsigned char)(rand() % 256);
	memcpy(h_DataA, hh_DataA, DATA_SIZE_INT*FRAMES);

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Grids and blocks
	///////////////////////////////////////////////////////////////////////////////////////////////////
	dim3 BLOCK_S((DATA_SIZE/THREAD_N)*(FRAMES/nstreams));

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef WARMUP
	//printf("Warm up...\n");
	cudaMemcpy(d_DataA+0*DATA_SIZE*4, hh_DataA+0*DATA_SIZE*4, DATA_SIZE_INT, cudaMemcpyHostToDevice);
	rgb2grayGPU<<<BLOCK_N, THREAD_N>>>(d_histo+0*BIN_COUNT, (unsigned int *)(d_DataA+0*DATA_SIZE*4), DATA_SIZE);
	cudaThreadSynchronize();
#endif
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stream management
	cudaStream_t stream[nstreams];
	for (i = 0; i < nstreams; ++i) cudaStreamCreate(&stream[i]);
	// Create CUDA event handles
	cudaEvent_t start_event,stop_event;
	cudaEventCreate(&start_event);
	cudaEventCreate(&stop_event);

for (x = 0; x < nreps; ++x){
	srand(rand());
	for(i = 0; i < 4*DATA_SIZE*FRAMES; i++)
		hh_DataA[i] = (unsigned char)(rand() % 256);

	memcpy(h_DataA, hh_DataA, DATA_SIZE_INT*FRAMES);
	////////////////////////////////////// TIME MEMCOPY HOST TO DEVICE /////////////////////////////////////////
		cudaEventRecord(start_event,0);
        cudaMemcpy(d_DataA, hh_DataA, DATA_SIZE_INT*FRAMES, cudaMemcpyHostToDevice);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&time_memcpy1,start_event,stop_event);
		TIME_MEMCPY1 += time_memcpy1;

	//////////////////////////////////////////// TIME KERNEL ////////////////////////////////////////////////////
		cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
		cudaEventRecord(start_event,0);
		rgb2grayGPU<<<DATA_SIZE*FRAMES/THREAD_N, THREAD_N>>>(d_histo, (unsigned int *)d_DataA, FRAMES*DATA_SIZE);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&time_kernel,start_event,stop_event);
		TIME_KERNEL += time_kernel;
		
	////////////////////////////////////// TIME MEMCOPY DEVICE TO HOST /////////////////////////////////////////
		cudaEventRecord(start_event,0);
        cudaMemcpy(h_histoGPU,d_histo,FRAMES*HISTOGRAM_SIZE,cudaMemcpyDeviceToHost);
		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&time_memcpy2,start_event,stop_event);
		TIME_MEMCPY2 += time_memcpy2;

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// NON-STREAMED EXECUTION //////////////////////////////////////////////
		cudaMemset(d_DataA, 0, DATA_SIZE_INT*FRAMES);
		cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
		cudaEventRecord(start_event,0);

		cudaMemcpy(d_DataA, hh_DataA, DATA_SIZE_INT*FRAMES, cudaMemcpyHostToDevice);

		rgb2grayGPU<<<DATA_SIZE*FRAMES/THREAD_N, THREAD_N>>>(d_histo, (unsigned int *)d_DataA, FRAMES*DATA_SIZE);

		cudaMemcpy(h_histoGPU1,d_histo,HISTOGRAM_SIZE*FRAMES,cudaMemcpyDeviceToHost);

		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&time_nostream,start_event,stop_event);
		TIME_nostream += time_nostream;

	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// STREAMED EXECUTION //////////////////////////////////////////////////
		cudaMemset(d_DataA, 0, DATA_SIZE_INT*FRAMES);
		cudaMemset(d_histo, 0, HISTOGRAM_SIZE*FRAMES);
		cudaEventRecord(start_event,0);

		for (i = 0; i < nstreams; ++i)	// As hh_DataA is unsigned char, we multiply by 4
			cudaMemcpyAsync(d_DataA+i*4*DATA_SIZE*FRAMES/nstreams, hh_DataA+i*4*DATA_SIZE*FRAMES/nstreams, DATA_SIZE_INT*FRAMES/nstreams, cudaMemcpyHostToDevice, stream[i]);

		for (i = 0; i < nstreams; ++i)
			rgb2grayGPU<<<BLOCK_S, THREAD_N, 0, stream[i]>>>(d_histo+i*BIN_COUNT*FRAMES/nstreams, (unsigned int *)(d_DataA+i*4*DATA_SIZE*FRAMES/nstreams), DATA_SIZE*FRAMES/nstreams);

       	for (i = 0; i < nstreams; ++i)
			cudaMemcpyAsync(h_histoGPU2+i*BIN_COUNT*FRAMES/nstreams, d_histo+i*BIN_COUNT*FRAMES/nstreams, HISTOGRAM_SIZE*FRAMES/nstreams, cudaMemcpyDeviceToHost, stream[i]);

		cudaEventRecord(stop_event,0);
		cudaEventSynchronize(stop_event);
		cudaEventElapsedTime(&time_stream,start_event,stop_event);
		TIME_stream += time_stream;

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// CPU computing
	///////////////////////////////////////////////////////////////////////////////////////////////////
	memset(h_histoCPU, 0, HISTOGRAM_SIZE*FRAMES);
	cudaThreadSynchronize();
	cudaEventRecord(start_event,0);
	for(i = 0; i < FRAMES; ++i){
		rgb2grayCPU(h_histoCPU+i*BIN_COUNT, (unsigned int *)h_DataA+i*DATA_SIZE, DATA_SIZE);
	}
	cudaEventRecord(stop_event,0);
	cudaEventSynchronize(stop_event);
	cudaEventElapsedTime(&elapsed_time,start_event,stop_event);
	ELAPSED_TIME += elapsed_time;
}

	printf("Device %s\tDATA_W %d\tFRAMES %d\tnstreams %d\n", device_properties.name, DATA_W, FRAMES, nstreams);

	printf("CPU-GPU %f\tKernel %f\tGPU-CPU %f\n", TIME_MEMCPY1/nreps, TIME_KERNEL/nreps, TIME_MEMCPY2/nreps);

	printf("GPU_no_stream %f\tGPU_stream %f ( %f / %f)\tCPU %f\n", TIME_nostream/nreps, TIME_stream/nreps, (TIME_MEMCPY1/nstreams+TIME_KERNEL+TIME_MEMCPY2/nstreams)/nreps, (TIME_MEMCPY1+TIME_MEMCPY2)/nreps, ELAPSED_TIME/nreps);

	printf("BLOCK_NS %d\tBLOCK_S %d\t", DATA_SIZE*FRAMES/THREAD_N, BLOCK_S.x);

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Comparing the results
	///////////////////////////////////////////////////////////////////////////////////////////////////
    sum_delta2 = 0;
    sum_ref2   = 0;
    L1norm2 = 0;
    for(int i = 0; i < BIN_COUNT*FRAMES; i++){
        sum_delta2 += fabs(h_histoGPU2[i] - h_histoGPU1[i]);
        sum_ref2   += fabs(h_histoGPU2[i]);
    }
    L1norm2 = sum_delta2 / sum_ref2;
    //printf("L1norm2: %E\n", L1norm2);
    printf((L1norm2 < 1e-6) ? "PASSED\n" : "FAILED\n");

	///////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	///////////////////////////////////////////////////////////////////////////////////////////////////
	for(int i = 0; i < nstreams; i++)
		cudaStreamDestroy(stream[i]);
	cudaEventDestroy(start_event);
	cudaEventDestroy(start_event);
	// Device memory
        cudaFree(d_DataA);
        cudaFree(d_histo);
	// Host memory
        cudaFreeHost(h_DataA);
        cudaFreeHost(hh_DataA);
        cudaFreeHost(h_histoCPU);
        cudaFreeHost(h_histoGPU1);
        cudaFreeHost(h_histoGPU2);
	cutDeleteTimer(hTimer);

}

///////////////////////////////////////////////////////////////////////////////////////////////////
