/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CUDA streams - JGL - 10/11/2010
// 
// Juan Gómez-Luna*, José María González-Linares#, José Ignacio Benavides* and Nicolás Guil#
// *University of Córdoba, #University of Málaga (Spain)
// http://www.ac.uma.es/~vip/
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Streamed RGB to grayscale
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// streams_RGB2gray_kernel.cu
// This file contains the GPU code used in Section 4.3.
// This code is a variation of histogram256_kernel.cu to compute grayscale histogram.
//
// Reference:
// @article{GomezLuna20121117,
//   title = "Performance models for asynchronous data transfers on consumer Graphics Processing Units",
//   journal = "Journal of Parallel and Distributed Computing",
//   volume = "72",
//   number = "9",
//   pages = "1117 - 1126",
//   year = "2012",
//   note = "<ce:title>Accelerators for High-Performance Computing</ce:title>",
//   issn = "0743-7315",
//   doi = "10.1016/j.jpdc.2011.07.011",
//   url = "http://www.sciencedirect.com/science/article/pii/S0743731511001468",
//   author = "Juan G\'omez-Luna and Jos\'e Mar\'ia Gonz\'alez-Linares and Jos\'e Ignacio Benavides and Nicol\'as Guil",
// }
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <cutil_inline.h>

#define BIN_COUNT 256
#define HISTOGRAM_SIZE (BIN_COUNT * sizeof(unsigned int))

////////////////////////////////////////////////////////////////////////////////
// GPU-specific definitions
////////////////////////////////////////////////////////////////////////////////
//Machine warp size
#define WARP_LOG2SIZE 5

//Warps in thread block for histogram256Kernel()
#define WARP_N 8

//Corresponding thread block size in threads for histogram256Kernel()
#define THREAD_N (WARP_N << WARP_LOG2SIZE)

//Total histogram size (in counters) per thread block for histogram256Kernel()
#define BLOCK_MEMORY (WARP_N * BIN_COUNT)

//Thread block count for histogram256Kernel()
#define BLOCK_N 64

//#define D_S 1584 // Pal
//#define D_S 396 // Half-pal
#define D_S 99 // Quarter-pal

////////////////////////////////////////////////////////////////////////////////
// Shortcut shared memory atomic addition functions
////////////////////////////////////////////////////////////////////////////////
__device__ inline void addData256(unsigned int *s_WarpHist, unsigned int data){
        atomicAdd(s_WarpHist + data, 1);
}

////////////////////////////////////////////////////////////////////////////////
// Main histogram calculation kernel
////////////////////////////////////////////////////////////////////////////////
static __global__ void rgb2grayGPU(unsigned int *d_Result, unsigned int *d_Data, int dataN){
    //Current global thread index
    const int    globalTid = blockIdx.x * blockDim.x + threadIdx.x;
    //Total number of threads in the compute grid
    const int   numThreads = blockDim.x * gridDim.x;
    const int H = (int)(blockIdx.x / D_S);

    //Shared memory storage for each warp
    __shared__ unsigned int s_Hist[BLOCK_MEMORY];

    //Current warp shared memory base
    const int warpBase = (threadIdx.x >> WARP_LOG2SIZE) * BIN_COUNT;

    //Clear shared memory buffer for current thread block before processing
    for(int pos = threadIdx.x; pos < BLOCK_MEMORY; pos += blockDim.x)
       s_Hist[pos] = 0;

    //Cycle through the entire data set, update subhistograms for each warp
    __syncthreads();

    for(int pos = globalTid; pos < dataN; pos += numThreads){
        unsigned int data4 = d_Data[pos];

        unsigned int B = (data4 >>  0) & 0xFFU;
        unsigned int G = (data4 >>  8) & 0xFFU;
        unsigned int R = (data4 >>  16) & 0xFFU;
                
        addData256(s_Hist + warpBase, (int)(0.30*R + 0.59*G + 0.11*B));

    }

    __syncthreads();
    //Merge per-warp histograms into per-block and write to global memory
    for(int pos = threadIdx.x; pos < BIN_COUNT; pos += blockDim.x){
        unsigned int sum = 0;

        for(int base = 0; base < BLOCK_MEMORY; base += BIN_COUNT)
            sum += s_Hist[base + pos] & 0x07FFFFFFU;
            atomicAdd(d_Result + H*BIN_COUNT + pos, sum);
    }
}

////////////////////////////////////////////////////////////////////////////////
